# @file edk2_db.py
# A class for interacting with a database implemented using json.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A class for interacting with a database implemented using json."""
import logging
import sqlite3
import time
import uuid
from typing import Any

from edk2toollib.database.tables import EnvironmentTable
from edk2toollib.database.tables.base_table import TableGenerator
from edk2toollib.uefi.edk2.path_utilities import Edk2Path

CREATE_JUNCTION_TABLE = """
CREATE TABLE IF NOT EXISTS junction (
    env TEXT,
    table1 TEXT,
    key1 TEXT,
    table2 TEXT,
    key2 TEXT
)
"""

CREATE_JUNCTION_INDEX = """
CREATE INDEX IF NOT EXISTS junction_idx
ON junction (env);
"""

class Edk2DB:
    """A SQLite3 database manager for a EDKII workspace.

    This class provides the ability to register parsers that will create / update tables in the database. This will
    create a SQLite datbase file that can be queried using any SQLite3 client. VSCode provides multiple extensions
    for viewing and interacting with the database. Queries can also be created and run in python using the sqlite3
    module that comes with python.

    Edk2DB can, and should, be used as a context manager to ensure that the database is closed properly. If
    not using as a context manager, the `db.connection.commit()` and `db.connection.close()` must be used to cleanly
    close the database.

    Attributes:
        connection (sqlite3.Connection): The connection to the database

    !!! note
        Edk2DB provides a table called `junction` that can be used to make associations between tables. It has the
        following schema: `env_id, table1, key1, table2, key2`.

    Example:
        ```python
        from edk2toollib.database.parsers import *
        table = "..."
        with Edk2DB(Path("path/to/db.db"), edk2path) as db:
            db.register(Parser1(), Parser2(), Parser3())
            db.parse()

            db.connection.execute("SELECT * FROM ?", table)
    """
    def __init__(self, db_path: str, pathobj: Edk2Path = None, **kwargs: dict[str,Any]):
        """Initializes the database.

        Args:
            db_path: Path to create or load the database from
            pathobj: Edk2Path object for the workspace
            **kwargs: None
        """
        self.pathobj = pathobj
        self.clear_parsers()
        self.connection = sqlite3.connect(db_path)

    def __enter__(self):
        """Enables the use of the `with` statement."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Enables the use of the `with` statement."""
        self.connection.commit()
        self.connection.close()

    def register(self, *parsers: 'TableGenerator') -> None:
        """Registers a one or more table generators.

        Args:
            *parsers: One or more instantiated TableGenerator object
        """
        for parser in parsers:
            self._parsers.append(parser)

    def clear_parsers(self) -> None:
        """Empties the list of registered table generators."""
        self._parsers = [EnvironmentTable()]

    def parse(self, env: dict) -> None:
        """Runs all registered table parsers against the database.

        !!! note
            To enable queries to differentiate between two parses, an environment table is always created if it does
            not exist, and a row is added for each call of this command.
        """
        self.connection.execute(CREATE_JUNCTION_TABLE)
        self.connection.execute(CREATE_JUNCTION_INDEX)
        id = str(uuid.uuid4().hex)

        # Create all tables
        for table in self._parsers:
            table.create_tables(self.connection.cursor())

        # Fill all tables
        for table in self._parsers:
            logging.debug(f"[{table.__class__.__name__}] starting...")
            t = time.time()
            table.parse(self.connection.cursor(), self.pathobj, id, env)
            self.connection.commit()
            logging.debug(f"Finished in {round(time.time() - t, 2)}")



# @file instaced_fv.py
# A module to run a table generator that uses a fdf and environment information to generate a table of information
# about instanced fvs where each row is a unique fv.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A module to generate a table containing fv information."""
import logging
import re
import sqlite3
from pathlib import Path

from edk2toollib.database.tables.base_table import TableGenerator
from edk2toollib.uefi.edk2.parsers.fdf_parser import FdfParser as FdfP
from edk2toollib.uefi.edk2.path_utilities import Edk2Path

CREATE_INSTANCED_FV_TABLE = """
CREATE TABLE IF NOT EXISTS instanced_fv (
    env INTEGER,
    fv_name TEXT,
    fdf TEXT,
    path TEXT
)
"""

INSERT_INSTANCED_FV_ROW = """
INSERT INTO instanced_fv (env, fv_name, fdf, path)
VALUES (?, ?, ?, ?)
"""

INSERT_JUNCTION_ROW = '''
INSERT INTO junction (env, table1, key1, table2, key2)
VALUES (?, ?, ?, ?, ?)
'''

class InstancedFvTable(TableGenerator):
    """A Table Generator that parses a single FDF file and generates a table containing FV information."""  # noqa: E501

    RULEOVERRIDE = re.compile(r'RuleOverride\s*=.+\s+(.+\.inf)', re.IGNORECASE)

    def __init__(self, *args, **kwargs):
        """Initialize the query with the specific settings."""


    def create_tables(self, db_cursor: sqlite3.Cursor) -> None:
        """Create the tables necessary for this parser."""
        db_cursor.execute(CREATE_INSTANCED_FV_TABLE)

    def parse(self, db_cursor: sqlite3.Cursor, pathobj: Edk2Path, env_id, env) -> None:
        """Parse the workspace and update the database."""
        self.pathobj = pathobj
        self.ws = Path(self.pathobj.WorkspacePath)
        self.env = env
        self.env_id = env_id
        self.dsc = self.env.get("ACTIVE_PLATFORM", None)
        self.fdf = self.env.get("FLASH_DEFINITION", None)
        self.arch = self.env["TARGET_ARCH"].split(" ")
        self.target = self.env["TARGET"]

        if self.dsc is None or self.fdf is None:
            logging.debug("DSC or FDF not found in environment. Skipping InstancedFvTable")
            return

        # Our DscParser subclass can now parse components, their scope, and their overrides
        fdfp = FdfP().SetEdk2Path(self.pathobj)
        fdfp.SetInputVars(self.env)
        fdfp.ParseFile(self.fdf)

        for fv in fdfp.FVs:

            inf_list = []  # Some INF's start with RuleOverride. We only need the INF
            for inf in fdfp.FVs[fv]["Infs"]:
                if inf.lower().startswith("ruleoverride"):
                    inf = InstancedFvTable.RULEOVERRIDE.findall(inf)[0]
                if Path(inf).is_absolute():
                    inf = str(Path(self.pathobj.GetEdk2RelativePathFromAbsolutePath(inf)))
                inf_list.append(Path(inf).as_posix())

            row = (self.env_id, fv, Path(self.fdf).name, self.fdf)
            db_cursor.execute(INSERT_INSTANCED_FV_ROW, row)

            for inf in inf_list:
                row = (self.env_id, "instanced_fv", fv, "inf", inf)
                db_cursor.execute(INSERT_JUNCTION_ROW, row)



# @file package_table.py
# A module to associate the packages in a workspace with the repositories they come from.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A module to generate a table containing information about a package."""
from pathlib import Path
from sqlite3 import Cursor

import git

from edk2toollib.database.tables.base_table import TableGenerator
from edk2toollib.uefi.edk2.path_utilities import Edk2Path

CREATE_PACKAGE_TABLE = """
CREATE TABLE IF NOT EXISTS package (
    name TEXT PRIMARY KEY,
    repository TEXT
)
"""

INSERT_PACKAGE_ROW = """
INSERT OR REPLACE INTO package (name, repository)
VALUES (?, ?)
"""
class PackageTable(TableGenerator):
    """A Table Generator that associates packages with their repositories."""
    def __init__(self, *args, **kwargs):
        """Initializes the Repository Table Parser.

        Args:
            args (any): non-keyword arguments
            kwargs (any): None

        """

    def create_tables(self, db_cursor: Cursor) -> None:
        """Create the table necessary for this parser."""
        db_cursor.execute(CREATE_PACKAGE_TABLE)

    def parse(self, db_cursor: Cursor, pathobj: Edk2Path, id: str, env: dict) -> None:
        """Glob for packages and insert them into the table."""
        try:
            repo = git.Repo(pathobj.WorkspacePath)
        except git.InvalidGitRepositoryError:
            return

        for file in Path(pathobj.WorkspacePath).rglob("*.dec"):
            pkg = pathobj.GetContainingPackage(str(file))
            containing_repo = "BASE"
            if repo:
                for submodule in repo.submodules:
                    if submodule.abspath in str(file):
                        containing_repo = submodule.name
                        break
            row = (pkg, containing_repo)
            db_cursor.execute(INSERT_PACKAGE_ROW, row)



# @file environment_table.py
# A module to run a table generator that creates or appends to a table with environment information."
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A module to run a table generator that creates or appends to a table with environment information."""
import datetime
import sqlite3

import git

from edk2toollib.database.tables.base_table import TableGenerator
from edk2toollib.uefi.edk2.path_utilities import Edk2Path

CREATE_ENV_TABLE_COMMAND = '''
CREATE TABLE IF NOT EXISTS environment (
    id TEXT PRIMARY KEY,
    date TEXT,
    version TEXT
);
'''

CREATE_ENV_VALUES_TABLE_COMMAND = '''
CREATE TABLE IF NOT EXISTS environment_values (
    id TEXT,
    key TEXT,
    value TEXT,
    FOREIGN KEY (id) REFERENCES environment(id)
);
'''

class EnvironmentTable(TableGenerator):
    """A Workspace parser that records import environment information for a given parsing execution."""  # noqa: E501
    def __init__(self, *args, **kwargs):
        """Initialize the query with the specific settings."""

    def create_tables(self, db_cursor: sqlite3.Cursor) -> None:
        """Create the tables necessary for this parser."""
        db_cursor.execute(CREATE_ENV_VALUES_TABLE_COMMAND)
        db_cursor.execute(CREATE_ENV_TABLE_COMMAND)

    def parse(self, db_cursor: sqlite3.Cursor, pathobj: Edk2Path, id, env) -> None:
        """Parses the environment and adds the data to the table."""
        dtime = datetime.datetime.now()

        try:
            version = git.Repo(pathobj.WorkspacePath).head.commit.hexsha
        except git.InvalidGitRepositoryError:
            version = "UNKNOWN"

        # Insert into environment table
        entry = (id,str(dtime),version,)
        db_cursor.execute("INSERT INTO environment (id, date, version) VALUES (?, ?, ?)", entry)

        # Insert into environment_values table
        data = [(id, key, value) for key, value in env.items()]
        db_cursor.executemany("INSERT INTO environment_values VALUES (?, ?, ?)", data)



# @file __init__.py
# The core classes and methods used to interact with the database portion of edk2-pytool-library
# This prevents needing to do deeply nested imports and can simply `from edk2toollib.database import``
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Core classes and methods used to interact with the database module inside edk2-pytool-library."""
from .edk2_db import Edk2DB  # noqa: F401



# @file inf_table.py
# A module to run a table generator that parses all INF files in the workspace and generates a table of information
# about each INF.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A module to run generate a table containing information about each INF in the workspace."""
import logging
import time
from pathlib import Path
from sqlite3 import Cursor

from joblib import Parallel, delayed

from edk2toollib.database.tables.base_table import TableGenerator
from edk2toollib.uefi.edk2.parsers.inf_parser import InfParser as InfP
from edk2toollib.uefi.edk2.path_utilities import Edk2Path

CREATE_INF_TABLE = '''
CREATE TABLE IF NOT EXISTS inf (
    path TEXT PRIMARY KEY,
    guid TEXT,
    library_class TEXT,
    package TEXT
);
'''

INSERT_JUNCTION_ROW = '''
INSERT INTO junction (env, table1, key1, table2, key2)
VALUES (?, ?, ?, ?, ?)
'''

INSERT_INF_ROW = '''
INSERT OR REPLACE INTO inf (path, guid, library_class, package)
VALUES (?, ?, ?, ?)
'''

class InfTable(TableGenerator):
    """A Table Generator that parses all INF files in the workspace and generates a table."""
    # TODO: Add phase, protocol, guid, ppi, pcd tables and associations once necessary
    def __init__(self, *args, **kwargs):
        """Initializes the INF Table Parser.

        Args:
            args (any): non-keyword arguments
            kwargs (any): keyword arguments described below

        Keyword Arguments:
            n_jobs (int): Number of files to run in parallel
        """
        self.n_jobs = kwargs.get("n_jobs", -1)

    def create_tables(self, db_cursor: Cursor) -> None:
        """Create the tables necessary for this parser."""
        db_cursor.execute(CREATE_INF_TABLE)

    def parse(self, db_cursor: Cursor, pathobj: Edk2Path, env_id: str, env: dict) -> None:
        """Parse the workspace and update the database."""
        ws = Path(pathobj.WorkspacePath)
        inf_entries = []

        start = time.time()
        files = list(ws.glob("**/*.inf"))
        files = [file for file in files if not file.is_relative_to(ws / "Build")]
        inf_entries = Parallel(n_jobs=self.n_jobs)(delayed(self._parse_file)(fname, pathobj) for fname in files)
        logging.debug(
            f"{self.__class__.__name__}: Parsed {len(inf_entries)} .inf files took; "
            f"{round(time.time() - start, 2)} seconds.")

        # Insert the data into the database
        for inf in inf_entries:
            row = (inf["PATH"], inf["GUID"], inf["LIBRARY_CLASS"], inf["PACKAGE"])
            db_cursor.execute(INSERT_INF_ROW, row)

            for library in inf["LIBRARIES_USED"]:
                row = (env_id, "inf", inf["PATH"], "library_class", library)
                db_cursor.execute(INSERT_JUNCTION_ROW, row)

            for source in inf["SOURCES_USED"]:
                source_path = (Path(inf["PATH"]).parent / source).as_posix()
                row = (env_id, "inf", inf["PATH"], "source", source_path)
                db_cursor.execute(INSERT_JUNCTION_ROW, row)

    def _parse_file(self, filename, pathobj) -> dict:
        inf_parser = InfP().SetEdk2Path(pathobj)
        inf_parser.ParseFile(filename)

        pkg = pathobj.GetContainingPackage(str(inf_parser.Path))
        path = Path(inf_parser.Path).as_posix()
        if pkg:
            path = path[path.find(pkg):]
        data = {}
        data["GUID"] = inf_parser.Dict.get("FILE_GUID", "")
        data["LIBRARY_CLASS"] = inf_parser.LibraryClass or None
        data["PATH"] = Path(path).as_posix()
        data["PHASES"] = inf_parser.SupportedPhases
        data["SOURCES_USED"] = inf_parser.Sources
        data["BINARIES_USED"] = inf_parser.Binaries
        data["LIBRARIES_USED"] = inf_parser.LibrariesUsed
        data["PROTOCOLS_USED"] = inf_parser.ProtocolsUsed
        data["GUIDS_USED"] = inf_parser.GuidsUsed
        data["PPIS_USED"] = inf_parser.PpisUsed
        data["PCDS_USED"] = inf_parser.PcdsUsed
        data["PACKAGE"] = pkg

        return data



# @file source_table.py
# A module to Parse all Source files and add them to the database.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A module to Parse all Source files and add them to the database."""
import logging
import re
import time
from pathlib import Path
from sqlite3 import Cursor

from joblib import Parallel, delayed

from edk2toollib.database.tables.base_table import TableGenerator
from edk2toollib.uefi.edk2.path_utilities import Edk2Path

SOURCE_FILES = ["*.c", "*.h", "*.cpp", "*.asm", "*.s", "*.nasm", "*.masm", "*.rs"]

CREATE_SOURCE_TABLE = '''
CREATE TABLE IF NOT EXISTS source (
    path TEXT UNIQUE,
    license TEXT,
    total_lines INTEGER,
    code_lines INTEGER,
    comment_lines INTEGER,
    blank_lines INTEGER
)
'''

INSERT_SOURCE_ROW = '''
INSERT OR REPLACE INTO source (path, license, total_lines, code_lines, comment_lines, blank_lines)
VALUES (?, ?, ?, ?, ?, ?)
'''


class SourceTable(TableGenerator):
    """A Table Generator that parses all c and h files in the workspace."""
    def __init__(self, *args, **kwargs):
        """Initializes the Source Table Parser.

        Args:
            args (any): non-keyword arguments
            kwargs (any): keyword arguments described below

        Keyword Arguments:
            n_jobs (int): Number of files to run in parallel
        """
        self.n_jobs = kwargs.get("n_jobs", -1)

    def create_tables(self, db_cursor: Cursor) -> None:
        """Create the tables necessary for this parser."""
        db_cursor.execute(CREATE_SOURCE_TABLE)

    def parse(self, db_cursor: Cursor, pathobj: Edk2Path, id: str, env: dict) -> None:
        """Parse the workspace and update the database."""
        ws = Path(pathobj.WorkspacePath)
        self.pathobj = pathobj

        start = time.time()
        files = []
        for src in SOURCE_FILES:
            files.extend(list(ws.rglob(src)))
        files = [file for file in files if not file.is_relative_to(ws / "Build")]
        src_entries = Parallel(n_jobs=self.n_jobs)(delayed(self._parse_file)(ws, filename) for filename in files)
        logging.debug(
            f"{self.__class__.__name__}: Parsed {len(src_entries)} files; "
            f"took {round(time.time() - start, 2)} seconds.")

        db_cursor.executemany(INSERT_SOURCE_ROW, src_entries)

    def _parse_file(self, ws, filename: Path) -> dict:
        """Parse a C file and return the results."""
        license = ""
        with open(filename, 'r', encoding='cp850') as f:
            for line in f.readlines():
                match = re.search(r"SPDX-License-Identifier:\s*(.*)$", line)  # TODO: This is not a standard format.
                if match:
                    license = match.group(1)
        return (
            self.pathobj.GetEdk2RelativePathFromAbsolutePath(filename.as_posix()),  # path
            license or "Unknown",  # license
            0,  # total_lines
            0,  # code_lines
            0,  # comment_lines
            0,  # blank_lines
        )



# @file base_table.py
# An interface for a parser that generates a sqlite3 table maintained by Edk2DB.
##
# Copyright (c) Microsoft Corporation
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""An interface for a parser that generates a sqlite3 table maintained by Edk2DB."""
import sqlite3

from edk2toollib.uefi.edk2.path_utilities import Edk2Path


class TableGenerator:
    """An interface for a parser that generates a sqlite3 table maintained by Edk2DB.

    Allows you to parse a workspace, file, etc, and load the contents into the database as rows in a table.

    Edk2Db provides a connection to a sqlite3 database and will commit any changes made during `parse` once
    the parser has finished executing and has returned. Review sqlite3 documentation for more information on
    how to interact with the database.
    """
    def __init__(self, *args, **kwargs):
        """Initialize the query with the specific settings."""

    def create_tables(self, db_cursor: sqlite3.Cursor) -> None:
        """Create the tables necessary for this parser."""
        raise NotImplementedError

    def parse(self, db_cursor: sqlite3.Cursor, pathobj: Edk2Path, id: str, env: dict) -> None:
        """Execute the parser and update the database."""
        raise NotImplementedError
